import sys
sys.path.append('../')
import pytest

from euclipy.core import *
from euclipy.measure import *
from euclipy.polygon import *
from euclipy.theorems import *
from euclipy.tools import *

def test_registry_measure_object():
    try:
        del Registry.instance
    except AttributeError:
        pass
    Segment([Point("A"), Point("B")]).measure.value = 1
    assert len(Registry().measures['SegmentMeasure']) == 1
    assert len(Registry().objects['Segment']) == 1


import sys
sys.path.append('../')
import pytest

from euclipy.core import *
from euclipy.measure import *
from euclipy.polygon import *
from euclipy.theorems import *
from euclipy.tools import *

def test_point_identity():
    assert Point("A") is Point("A")

def test_point_inquality():
    assert Point("A") is not Point("B")

def test_segment_identity():
    assert Segment([Point("A"), Point("B")]) is Segment([Point("A"), Point("B")])

def test_segment_inequality():
    assert Segment([Point("A"), Point("B")]) is not Segment([Point("B"), Point("C")])

def test_segment_measure_identity():
    assert SegmentMeasure(1) is SegmentMeasure(1)

def test_segment_measure_inequality():
    assert SegmentMeasure(1) is not SegmentMeasure(2)

def test_undefined_segment_measure():
    assert SegmentMeasure().value == None

def test_angle_measure_identity():
    assert AngleMeasure(60) is AngleMeasure(60)

def test_angle_measure_inequality():
    assert AngleMeasure(60) is not AngleMeasure(90)

def test_undefined_angle_measure():
    assert AngleMeasure().value == None

def test_dummy():
    assert True


from collections import defaultdict
import pprint
import warnings
import sympy
import sympy.core.function

class Registry:
    '''Singleton registry of:
        - Geometry instances (GeometricObjects and GeometricMeasures)
        - Theorems
    '''
    def __new__(cls):
        '''objects and measures are dictionaries with:
            - keys: registry key
            - values: dictionaries mapping labels to registered objects/measures
        '''
        if not hasattr(cls, 'instance'):
            cls.instance = super(Registry, cls).__new__(cls)
            cls.instance.objects = defaultdict(dict)
            cls.instance.measures = defaultdict(dict)
            cls.instance.theorems = defaultdict(list)
            cls.instance.auto_label_counter = defaultdict(int)
        return cls.instance
   
    def add_object(self, geometric_object):
        self.objects[geometric_object._registry_key][geometric_object.label] = geometric_object

    def add_measure(self, measure):
        self.measures[measure._registry_key][measure.label] = measure
    
    def remove_measure(self, measure):
        self.measures[measure._registry_key].pop(measure.label, None)

    def add_theorem(self, theorem, applies_to):
        self.theorems[applies_to].append(theorem)
            
    def get_auto_label(self, geometry):
        self.auto_label_counter[geometry._registry_key] += 1
        return f'{geometry._label_prefix}{self.auto_label_counter[geometry._registry_key]}'

    def find_object(self, registry_key, label):
        try:
            return self.objects[registry_key][label]
        except KeyError:
            return None

    def search_measure_by_class_and_value(self, measure_cls, value):
        try:
            return [m for m in self.measures[measure_cls._registry_key].values() if m.value == value][0]
        except IndexError:
            return None

    def search_measure_by_label(self, label):
        try:
            return[m for key in self.measures for m in self.measures[key].values() if m.label == label][0]
        except IndexError:
            return None

    def search_polygon(self, registry_key, points: list):
        '''Find polygon that shares points, regardless of point order
        '''
        try:
            matches = [p for p in iter(self.objects[registry_key].values()) if set(p.points) == set(points)]
            return matches[0] if matches else None
        except KeyError:
            return None

    def print_registry(self):
        pprint.pprint(self.objects)
        pprint.pprint(self.measures)
        pprint.pprint(self.theorems)

class Geometry:
    @classmethod
    @property
    def _registry_key(cls):
        return cls.__name__

class Expressions:
    '''
    Zero-valued expressions derived from theorems
    '''
    def __new__(cls):
        '''expressions is a list containing zero-valued expressions derived from theorems.
        The expressions are constructed in sympy. Symbols representing measures use the label of the corresponding measures.:
        '''
        if not hasattr(cls, 'instance'):
            cls.instance = super(Expressions, cls).__new__(cls)
            cls.instance.expressions = list()
        return cls.instance

    @staticmethod
    def measure_symbol(measure):
        return sympy.Symbol(measure.label)

    def add_expression(self, expression):
        original_expression = expression
        for s in expression.free_symbols:
            m = Registry().search_measure_by_label(s.name)
            if m.value:
                expression = expression.subs(s, m.value)
        print(expression)
        if expression == 0:
            warnings.warn(f'{original_expression} contained no unknowns', stacklevel=3)
        else:
            self.expressions.append(expression)

    def substitute(self, var, sol):
        for i, e in enumerate(self.expressions):
            self.expressions[i] = e.subs(var, sol)
    
    def solve(self):
        # Find systems with numbers of equations equal to the number of free variables and solve them, using sympy
        solution = sympy.solve(self.expressions, dict=True)
        # For all solutions with no free symbols, substitute the answer into the measure
        for system in solution:
            for value in system.values():
                if sympy.core.function._coeff_isneg(value):
                    continue
                else:
                    for var, sol in system.items():
                        if not sol.free_symbols:
                            # Find the measure that corresponds to the variable
                            var_measure = Registry().search_measure_by_label(var.name)
                            # Set the measure value to the solution
                            var_measure.value = sol
                            self.substitute(var, sol)
        self.expressions = [e for e in self.expressions if e != 0]


import sys
sys.path.append("./")

from collections import defaultdict
from euclipy.geometric_objects import Shape, Point, Segment, Angle
from euclipy.core import Registry

class Polygon(Shape):
    def __new__(cls, points: list):
        entry = Registry().search_polygon(cls._registry_key, points)
        points = cls.translate_shape_points(points)
        label = '-'.join([p.label for p in points])
        if entry is None:
            instance = super().__new__(cls, label)
            instance.points = points
        else:
            entry_label = '-'.join([p.label for p in entry.points])
            if label == entry_label:
                instance = entry
            else:
                raise Exception #TODO: Create custom exception
        instance.edges = [Segment(set((points + points)[i:i+2])) for i in range(len(points))]
        instance.angles = [Angle(list(reversed((points + points)[i:i+3]))) for i in range(len(points))]
        return instance

    def angle_at_vertex(self, vertex: Point) -> Angle:
        try:
            return [a for a in self.angles if a.vertex() == vertex][0]
        except IndexError:
            return None

    def unknown_angles(self):
        return [a for a in self.angles if a.measure.value is None]

    def known_angles(self):
        return [a for a in self.angles if a.measure.value is not None]

    def unknown_segments(self):
        return [s for s in self.segments if s.measure.value is None]

    def known_segments(self):
        return [s for s in self.segments if s.measure.value is not None]

    @staticmethod
    def translate_shape_points(points: list) -> list:
        '''Reorder points starting with the lexically first one, but preserving order otherwise
        For example [C, B, A] would be reordered as [A, C, B]
        '''
        point_labels = [p.label for p in points]
        lexical_first_loc = point_labels.index(min(point_labels))
        return points[lexical_first_loc:] + points[:lexical_first_loc]

class Triangle(Polygon):
    def __new__(cls, points: list):
        '''Points must be ordered in a clockwise motion.
        '''
        assert len(points) == 3
        return super().__new__(cls, points)

    def congruent_sides(self) -> list:
        side_map = defaultdict(list)
        for e in self.edges:
            side_map[e.measure].append(e)
        try:
            return [group for group in side_map.values() if len(group) > 1][0]
        except IndexError:
            return []

    def congruent_angles(self) -> list:
        angle_map = defaultdict(list)
        for a in self.angles:
            angle_map[a.measure].append(a)
        try:
            return [group for group in angle_map.values() if len(group) > 1][0]
        except IndexError:
            return []
    
    def dummy_fcn(self):
        return "Dummy"

    def is_right_triangle(self):
        try:
            right_angle = self.angles[[a.measure.value for a in self.angles].index(90)]
            hypotenuse = 0
            right_sides = []
            for edge in self.edges:
                if Angle.vertex(right_angle) not in edge.endpoints:
                    hypotenuse = edge
                else:
                    right_sides.append(edge)
            right_sides.append(hypotenuse)
            return right_sides
        except ValueError:
            return False


import sympy
import sys
sys.path.append("./")

from euclipy.core import Registry, Expressions
from euclipy.polygon import Triangle
from euclipy.tools import pairs_in_iterable


def triangle_sum_theorem(triangle:Triangle) -> bool:
    Expressions().add_expression(sum([a.measure.symbol for a in triangle.angles]) - 180)
    return True

Registry().add_theorem(triangle_sum_theorem, 'Triangle')

def pythagorean_theorem(triangle:Triangle) -> bool:
    is_right_triangle = Triangle.is_right_triangle(triangle)
    if is_right_triangle is not False:
        print(is_right_triangle)
        Expressions().add_expression(is_right_triangle[0].measure.symbol**2 + is_right_triangle[1].measure.symbol**2 - is_right_triangle[2].measure.symbol**2)
        return True

Registry().add_theorem(pythagorean_theorem, 'Triangle')

def isosceles_triangle_theorem(triangle: Triangle) -> bool:
    theorem_applied = False
    congruent_sides = triangle.congruent_sides()
    if congruent_sides:
        for side_pairs in pairs_in_iterable(congruent_sides):
            vertex = side_pairs[0].common_point_with(side_pairs[1])
            angles = [a for a in triangle.angles if a.vertex() != vertex]
            angles[0].measure.set_equal_to(angles[1].measure)
        theorem_applied = True
    congruent_angles = triangle.congruent_angles()
    if congruent_angles:
        for angle_pairs in pairs_in_iterable(congruent_angles):
            sides = [side for angle in angle_pairs for side in triangle.edges if angle.vertex() not in side.endpoints]
            sides[0].measure.set_equal_to(sides[1].measure)
        theorem_applied = True
    return theorem_applied

Registry().add_theorem(isosceles_triangle_theorem, 'Triangle')



import sys
sys.path.append("./")

from euclipy.core import Geometry, Registry, Expressions

class GeometricMeasure(Geometry):
    def __new__(cls, value=None) -> None:
        '''If value is None, then the measure is not yet quantified.
        '''
        # If value of measure is defined, then search registry for a measure of that type and value exists
        if value is not None:
            match = Registry().search_measure_by_class_and_value(cls, value)
            if match:
                return match
        # Either value is None or value has no match in Registry, so create new meesure instance
        cls.instance = super().__new__(cls)
        cls.instance._value = value
        cls.instance.measured_objects = set()
        cls.instance.label = Registry().get_auto_label(cls.instance)
        Registry().add_measure(cls.instance)
        return cls.instance
    
    def __repr__(self) -> str:
        return f'{self._registry_key}({self.label}={self.value})'

    def _add_measured_object(self, measured_object) -> None:
        self.measured_objects.add(measured_object)

    def set_equal_to(self, other_measure):
        if self is not other_measure:
            for measured_object in other_measure.measured_objects.union(self.measured_objects):
                measured_object.measure = self
            if self.value is not None and other_measure.value is not None and self.value != other_measure.value:
                raise ValueError
            else:
                self._value = self.value or other_measure.value
            self.measured_objects = self.measured_objects.union(other_measure.measured_objects)
            #TODO: In expressions, substitute other_measure.label with self.label
            Registry().remove_measure(other_measure)

    @property
    def symbol(self):
        return Expressions.measure_symbol(self)
    
    @property
    def value(self):
        return self._value
    
    @value.setter
    def value(self, new_value):
        if new_value is not None:
            match = Registry().search_measure_by_class_and_value(type(self), new_value)
            if match:
                match.set_equal_to(self)
                #TODO: Check if expressions contain self.label and replace with new_value
                return
        self._value = new_value

class SegmentMeasure(GeometricMeasure):
    _label_prefix = 's'
    def __new__(cls, value=None) -> None:
        return super().__new__(cls, value)

class AngleMeasure(GeometricMeasure):
    _label_prefix = 'a'
    def __new__(cls, value=None) -> None:
        return super().__new__(cls, value)


import sys
sys.path.append('../')
import pytest

from euclipy.core import *
from euclipy.measure import *
from euclipy.polygon import *
from euclipy.tools import *
from euclipy.theorems import *

def test_sum_theorem():
    try:
        del Registry.instance
    except AttributeError:
        pass
    T1 = Triangle([Point("A"), Point("B"), Point("C")])
    T1.angles[0].measure.value = 30
    T1.angles[1].measure.value = 60
    triangle_sum_theorem(T1)
    Expressions().solve()
    assert T1.angles[2].measure.value == 90

def test_isosceles_triangle_theorem():
    try:
        del Registry.instance
    except AttributeError:
        pass
    T1 = Triangle([Point("A"), Point("B"), Point("C")])
    T1.angles[0].measure.value = 45
    T1.angles[1].measure.value = 45
    T1.edges[2].measure.value = 1
    isosceles_triangle_theorem(T1)
    Expressions().solve()
    assert T1.edges[0].measure is T1.edges[2].measure

def test_pythagorean_theorem():
    try:
        del Registry.instance
    except AttributeError:
        pass
    T1 = Triangle([Point("A"), Point("B"), Point("C")])
    T1.angles[2].measure.value = 90
    T1.edges[0].measure.value = 3
    T1.edges[1].measure.value = 5
    pythagorean_theorem(T1)
    Expressions().solve()
    assert T1.edges[2].measure.value == 4
    
test_sum_theorem()
test_isosceles_triangle_theorem()
test_pythagorean_theorem()



import sys
sys.path.append("./")

from euclipy.core import Geometry, Registry
from euclipy.measure import SegmentMeasure, AngleMeasure

class GeometricObject(Geometry):
    def __new__(cls, label):
        entry = Registry().find_object(cls._registry_key, label)
        if entry is None:
            cls.instance = super().__new__(cls)
            cls.instance.label = label
            Registry().add_object(cls.instance)
            return cls.instance
        return entry

    def __repr__(self) -> str:
        return f'{self._registry_key}({self.label})'

    def create_measure_if_unmeasured(self) -> None:
        assert hasattr(self, '_measure_class')
        if not hasattr(self, 'measure'):
            self.measure = self._measure_class()
            self.measure._add_measured_object(self)

class Point(GeometricObject):

    def __new__(cls, label):
        return super().__new__(cls, label)

class Segment(GeometricObject):
    _measure_class = SegmentMeasure

    def __new__(cls, endpoints: set):
        label = '-'.join(sorted([p.label for p in endpoints]))
        instance = super().__new__(cls, label)
        instance.create_measure_if_unmeasured()
        instance.endpoints = endpoints
        return instance

    def __repr__(self) -> str:
        return f'{self._registry_key}({self.label} | {self.measure})'

    def common_point_with(self, segment) -> Point:
        common_points = self.endpoints.intersection(segment.endpoints)
        if len(common_points) != 2:
            try:
                return common_points.pop()
            except KeyError:
                return None
        else:
            raise ValueError

class Angle(GeometricObject):
    _measure_class = AngleMeasure

    def __new__(cls, points: list):
        '''Points must be ordered such that the angle represents the clockwise motion from the first defined segment to the second defined segment.
        For example, if points = [A, B, C], then the angle is the clockwise motion from Segment(AB) to Segment(BC).
        '''
        label = '-'.join([p.label for p in points])
        instance = super().__new__(cls, label)
        instance.create_measure_if_unmeasured()
        instance.points = points
        return instance

    def __repr__(self) -> str:
        return f'{self._registry_key}({self.label} | {self.measure})'

    def vertex(self) -> Point:
        return self.points[1]

class Shape(GeometricObject):
    def __new__(cls, label):
        return super().__new__(cls, label)


from setuptools import setup, find_packages
 
classifiers = [
  'Development Status :: 5 - Production/Stable',
  'Intended Audience :: Education',
  'Operating System :: Microsoft :: Windows :: Windows 10',
  'License :: OSI Approved :: MIT License',
  'Programming Language :: Python :: 3'
]

setup(
  name='euclipy',
  version='0.2.13',
  description='A Euclidean Geometry based library',
  long_description=open('README.md').read(),
  long_description_content_type='text/markdown',
  url='',
  project_urls={
            #"Bug Tracker": "",
            "Documentation": "https://github.com/joshuavaron/euclipy/blob/main/doc/documentation.md",
            "Source Code": "https://github.com/joshuavaron/euclipy",
        },
  author='Joshua Varon',    
  author_email='32440072+joshuavaron@users.noreply.github.com',
  license='MIT', 
  classifiers=classifiers,
  keywords='geometry, math',
  packages=find_packages(include=['euclipy']),
  install_requires=['sympy'] 
)
